/*-----------------------------------------------------------------------------
 * Name:    LED_EZ_LPC1788.c
 * Purpose: LED interface for Embedded Artists LPC1788-32 Developers Kit
 * Rev.:    1.0.1
 *----------------------------------------------------------------------------*/

/* Copyright (c) 2013 - 2017 ARM LIMITED

   All rights reserved.
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   - Neither the name of ARM nor the names of its contributors may be used
     to endorse or promote products derived from this software without
     specific prior written permission.
   *
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS AND CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.
   ---------------------------------------------------------------------------*/


#include "Driver_I2C.h"
#include "Board_LED.h"

#ifndef LED_I2C_PORT
#define LED_I2C_PORT    0               /* I2C Port number                    */
#endif

#define LED_I2C_ADDR    0x60            /* 7-bit I2C address                  */

/* I2C Driver */
#define _I2C_Driver_(n)  Driver_I2C##n
#define  I2C_Driver_(n) _I2C_Driver_(n)
extern ARM_DRIVER_I2C    I2C_Driver_(LED_I2C_PORT);
#define ptrI2C         (&I2C_Driver_(LED_I2C_PORT))

#define LED_COUNT (8)

/* PCA9532 registers */
#define PCA9532_LS2_REG      ((uint8_t)(0x08))
#define PCA9532_LS3_REG      ((uint8_t)(0x09))

/**
  \fn          int32_t LED_Initialize (void)
  \brief       Initialize LEDs
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t LED_Initialize (void) {
  uint8_t  data[2];

  if (ptrI2C->Initialize (NULL) == ARM_DRIVER_OK)
  {
    ptrI2C->PowerControl (ARM_POWER_FULL);
    ptrI2C->Control      (ARM_I2C_BUS_SPEED, ARM_I2C_BUS_SPEED_FAST);
    ptrI2C->Control      (ARM_I2C_BUS_CLEAR, 0);
  }

  data[0] = PCA9532_LS2_REG;
  data[1] = 0x00;
  ptrI2C->MasterTransmit (LED_I2C_ADDR, data, 2, false);
  while (ptrI2C->GetStatus().busy);

  data[0] = PCA9532_LS3_REG;
  data[1] = 0x00;
  ptrI2C->MasterTransmit (LED_I2C_ADDR, data, 2, false);
  while (ptrI2C->GetStatus().busy);

  return 0;
}

/**
  \fn          int32_t LED_Uninitialize (void)
  \brief       De-initialize LEDs
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t LED_Uninitialize (void) {
  uint8_t  data[2];

  data[0] = PCA9532_LS2_REG;
  data[1] = 0x00;
  ptrI2C->MasterTransmit (LED_I2C_ADDR, data, 2, false);
  while (ptrI2C->GetStatus().busy);

  data[0] = PCA9532_LS3_REG;
  data[1] = 0x00;
  ptrI2C->MasterTransmit (LED_I2C_ADDR, data, 2, false);
  while (ptrI2C->GetStatus().busy);

  return 0;
}

/**
  \fn          int32_t LED_On (uint32_t num)
  \brief       Turn on requested LED
  \param[in]   num  LED number
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t LED_On (uint32_t num) {
  uint8_t  val;
  uint8_t  data[2];
  uint8_t  reg = (num % 2)? PCA9532_LS3_REG : PCA9532_LS2_REG;
  int32_t retCode = 0;

  if (num < LED_COUNT) {
    if(num % 2) num--;

    data[0]  = reg;
    ptrI2C->MasterTransmit (LED_I2C_ADDR, data, 1, true);
    while (ptrI2C->GetStatus().busy);
    ptrI2C->MasterReceive (LED_I2C_ADDR, &val, 1, false);
    while (ptrI2C->GetStatus().busy);

    data[0]  = reg;
    data[1]  = (1 << num) | val;
    ptrI2C->MasterTransmit (LED_I2C_ADDR, data, 2, false);
    while (ptrI2C->GetStatus().busy);
  }
  else {
    retCode = -1;
  }

  return retCode;
}

/**
  \fn          int32_t LED_Off (uint32_t num)
  \brief       Turn off requested LED
  \param[in]   num  LED number
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t LED_Off (uint32_t num) {
  uint8_t  val;
  uint8_t  data[2];
  uint8_t  reg = (num % 2)? PCA9532_LS3_REG : PCA9532_LS2_REG;
  int32_t retCode = 0;

  if (num < LED_COUNT) {
    if(num % 2) num--;

    data[0]  = reg;
    ptrI2C->MasterTransmit (LED_I2C_ADDR, data, 1, true);
    while (ptrI2C->GetStatus().busy);
    ptrI2C->MasterReceive (LED_I2C_ADDR, &val, 1, false);
    while (ptrI2C->GetStatus().busy);

    data[0]  = reg;
    val     &= ~(1 << num);
    data[1]  = val;
    ptrI2C->MasterTransmit (LED_I2C_ADDR, data, 2, false);
    while (ptrI2C->GetStatus().busy);
  }
  else {
    retCode = -1;
  }

  return retCode;
}

/**
  \fn          int32_t LED_SetOut (uint32_t val)
  \brief       Write value to LEDs
  \param[in]   val  value to be displayed on LEDs
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t LED_SetOut (uint32_t val) {
  uint32_t n;

  for (n = 0; n < LED_COUNT; n++) {
    if (val & (1 << n)) LED_On (n);
    else                LED_Off(n);
  }

  return 0;
}

/**
  \fn          uint32_t LED_GetCount (void)
  \brief       Get number of LEDs
  \return      Number of available LEDs
*/
uint32_t LED_GetCount (void) {

  return LED_COUNT;
}
