/*------------------------------------------------------------------------------
 * MDK Middleware - Component ::USB:Device
 * Copyright (c) 2004-2016 ARM Germany GmbH. All rights reserved.
 *------------------------------------------------------------------------------
 * Name:    USBD_User_ADC_0.c
 * Purpose: USB Device Audio Device Class (ADC) User module
 *          Template for USB Audio Device Class (ADC)
 *          mono speaker driven by Digital-to-Analog Converter
 * Rev.:    V6.3.0
 *----------------------------------------------------------------------------*/
/**
 * \addtogroup usbd_adcFunctions
 *
 * USBD_User_ADC_0.c implements the application specific
 * functionality of the Audio class and is used to create an USB Audio device
 * containing mono speaker driven by Digital-to-Analog Converter.
 *
 * The implementation uses settings from USBD_Config_ADC_0.h file.
 *
 */
 
 
//! [code_USBD_User_ADC]
 
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include "rl_usb.h"
 
#include "LPC17xx.h"
 
#include ".\RTE\USB\USBD_Config_ADC_0.h"
 
// Imported settings from USBD_Config_ADC_0.h file
#define  PLAYBACK_AVAILABLE                 (USBD_ADC0_EP_ISO_OUT_EN != 0U)
#define  PLAYBACK_CHANNELS                  (USBD_ADC0_OUT_CH_NUM)
#define  PLAYBACK_FREQ                      (USBD_ADC0_OUT_TSAM_FREQ)
#define  PLAYBACK_SAMPLE_SIZE               (USBD_ADC0_OUT_BSUBFRAMESIZE)
#define  PLAYBACK_USB_BUFFER_SIZE           (USBD_ADC0_OUT_BUF_SIZE)
#define  PLAYBACK_AUDIO_BUFFER_SIZE         (USBD_ADC0_OUT_BUF_SIZE / 8U)
 
#define  RECORD_AVAILABLE                   (USBD_ADC0_EP_ISO_IN_EN != 0U)
 
// Supported configuration check
#if     (PLAYBACK_AVAILABLE != 0U)
#if     (PLAYBACK_CHANNELS  != 1U)
#error   Playback channel configuration mono only supported!
#endif
#if     (PLAYBACK_SAMPLE_SIZE != 2U)
#error   Playback channel resolutions of 16 bits only supported!
#endif
#else
#error   Playback channel not configured!
#endif
#if     (RECORD_AVAILABLE != 0U)
#error   Recording channel not supported!
#endif
 
extern int32_t  DAC_Initialize  (void);
extern int32_t  DAC_Uninitialize(void);
extern int32_t  DAC_SetValue    (int32_t val);
 
static uint8_t  play_freq_sync_cnt;
static bool     play_active;
static bool     play_mute;
static uint16_t play_volume;
 
 
// Timer IRQ happening at playback frequency
void TIMER0_IRQHandler (void) {
  uint32_t samples_available;
  int32_t  scaled_sample;
  uint16_t sample;
 
  if (play_active == true) {
    if (USBD_ADC_ReadSamples (0U, (void *)(&sample), 1U) == 1U) {
      if (play_mute == false) {
        scaled_sample   = (0x8000U + sample) & 0xFFFFU;
        scaled_sample  *= play_volume;
        scaled_sample >>= 8U;
        DAC_SetValue (scaled_sample);
      }
    }

    if (play_freq_sync_cnt++ == 0U) {
      // Handling of playback Timer period compensation
      // Every 256 played samples frequencies of the Timer and USB are compared
      // and Timer frequency is increased or decreased to correspond to USB frequency
      samples_available = USBD_ADC_ReceivedSamplesAvailable (0U);
      if (samples_available != 0U) {    // There are playback samples available
        if (samples_available <= PLAYBACK_AUDIO_BUFFER_SIZE) {
          // Lengthen Timer0 period if USB is slower than Timer0 produced frequency
          LPC_TIM0->MR0 += 1U;
        }
        if (samples_available >= (PLAYBACK_USB_BUFFER_SIZE - PLAYBACK_AUDIO_BUFFER_SIZE)) {
          // Shorten Timer0 period if USB is faster than Timer0 produced frequency
          LPC_TIM0->MR0 -= 1U;
        }
      } else {                          // No playback samples available
        play_active = false;            // Disable sample playback in Timer IRQ
        LPC_TIM0->TCR = 0U;             // Disable Timer0
      }
    }
  }

  LPC_TIM0->IR = 1U;                    // Clear Timer0 Interrupt
}
 
// Called during USBD_Initialize to initialize the USB ADC class instance.
void USBD_ADC0_Initialize (void) {

  // Initialize variables
  play_freq_sync_cnt = 0U;
  play_active        = false;
  play_mute          = false;
 
  SystemCoreClockUpdate ();             // Update SystemCoreClock value

  DAC_Initialize ();                    // Initialize DAC for speaker output
 
  // Set stereo speakers volume range to 0 to 256 with step 4, initially positioned at 50 %
  USBD_ADC_SpeakerSetVolumeRange (0U, 0U, 0U, 256U, 4U, 128U);
  play_volume = 128U;

  // Setup Timer0: Periodic Interrupt at playback frequency
  LPC_SC->PCONP |=  1U << 1U;           // Enable power to Timer0
  LPC_TIM0->MR0  = (SystemCoreClock/PLAYBACK_FREQ) - 1U;        // Timer0 Match Value 0
  LPC_TIM0->MCR  =  3U;                 // Interrupt and Reset on MR0
  NVIC_EnableIRQ(TIMER0_IRQn);          // Enable Timer0 Interrupt
}
 
// Called during USBD_Uninitialize to de-initialize the USB ADC class instance.
void USBD_ADC0_Uninitialize (void) {
  DAC_Uninitialize ();
}
 
// Callback function called when speaker activity (interface) setting changed event.
// \param[in]     active        activity status.
void USBD_ADC0_SpeakerStatusEvent (bool active) {
}
 
// Callback function called when speaker mute setting changed event.
// \param[in]     ch            channel index.
//                                - value 0: master channel
//                                - value 1: left speaker (in stereo mode)
//                                - value 2: right speaker (in stereo mode)
// \param[in]     cur           current setting.
void USBD_ADC0_SpeakerMuteEvent (uint8_t ch, bool cur) {
  play_mute = cur;
}
 
// Callback function called when speaker volume setting changed event.
// \param[in]     ch            channel index.
//                                - value 0: master channel
//                                - value 1: left speaker (in stereo mode)
//                                - value 2: right speaker (in stereo mode)
// \param[in]     cur           current setting.
void USBD_ADC0_SpeakerVolumeEvent (uint8_t ch, uint16_t cur) {
  play_volume = cur;
}
 
// Callback function called when microphone activity (interface) setting changed event.
// \param[in]     active        activity status.
void USBD_ADC0_MicrophoneStatusEvent (bool active) {
}
 
// Callback function called when microphone mute setting changed event.
// \param[in]     ch            channel index.
//                                - value 0: master channel
//                                - value 1: left microphone (in stereo mode)
//                                - value 2: right microphone (in stereo mode)
// \param[in]     cur           current setting.
void USBD_ADC0_MicrophoneMuteEvent (uint8_t ch, bool cur) {
}
 
// Callback function called when microphone volume setting changed event.
// \param[in]     ch            channel index.
//                                - value 0: master channel
//                                - value 1: left microphone (in stereo mode)
//                                - value 2: right microphone (in stereo mode)
// \param[in]     cur           current setting.
void USBD_ADC0_MicrophoneVolumeEvent (uint8_t ch, uint16_t cur) {
}
 
// Callback function called when data in buffer for speaker samples reaches from empty to half-full.
void USBD_ADC0_ReceivedSamples (void) {
  play_active = true;                   // Play samples in Timer IRQ
  if (LPC_TIM0->TCR == 0U) {            // If Timer0 is not active
    LPC_TIM0->TCR = 3U;                 // Reset Timer0
    LPC_TIM0->TCR = 1U;                 // Enable Timer0
  }
}
 
//! [code_USBD_User_ADC]
